#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include <exception>
#include <stdexcept>
#include <string>
#include <vector>

#include "datetime.h"

#ifdef ZERO
#undef ZERO
#endif
#define ZERO 0

#ifdef FALSE
#undef FALSE
#endif
#define FALSE ZERO

#ifdef TRUE
#undef TRUE
#endif
#define TRUE !FALSE

// #ifdef NULL
// #undef NULL
// #endif
// #define NULL ((void*) 0)

#ifdef NILP
#undef NILP
#endif
#define NILP ((void*) 0)

#define TEXT_NONSENSE "!§$§&/(/()a3f54"
#define NUMERIC_NONSENSE (-1)

#define TEXT_BAD_DATE "bad Date"
#define DOT "."
#define DATE_DELIMITER_US "-"
#define DATE_DELIMITER_DE "."

#define TEXT_BAD_TIME "bad Time"
#define TIME_DELIMITER ":"

#define SPACE " "
#define TEXT_BAD_DATETIME "bad DateTime"

#define DIGIT_0 "0"
#define DIGIT_1 "1"
#define DIGIT_2 "2"
#define DIGIT_3 "3"
#define DIGIT_4 "4"
#define DIGIT_5 "5"
#define DIGIT_6 "6"
#define DIGIT_7 "7"
#define DIGIT_8 "8"
#define DIGIT_9 "9"

static std::vector<std::string> split(std::string S, std::string to_find)
{
   int l;
   int n;
   int pos;
   std::string T;
   std::string U;
   std::vector<std::string>::iterator it;
   std::vector<std::string> V;

   if (to_find == "") return(V);
   
   pos = S.find(to_find);

   // to_find not found!
   if (pos == std::string::npos)
   {
      V.resize(1);
      it = V.begin();
      *it = S;
      return(V);
   }

   // Here S contains to_find!
   n = 0;
   T = S;
   while (pos != std::string::npos)
   {
      U = T.substr(0, pos);
      
      l = T.length() - (pos + to_find.length());
      T = T.substr(pos + to_find.length(), l);
      
      pos = T.find(to_find);
      n++;
   }
   U = T;
   // Here is (n >= 1)!

   V.resize(n + 1);
   it = V.begin();
   
   // Here S contains to_find!
   T = S;
   pos = S.find(to_find);
   while (pos != std::string::npos)
   {
      U = T.substr(0, pos);
      *it = U; it++;
      
      l = T.length() - (pos + to_find.length());
      T = T.substr(pos + to_find.length(), l);
      
      pos = T.find(to_find);
   }
   U = T;
   *it = U; it++;

   return(V);
};

static int InterpretDecimalDigit(std::string S)
{
   int i;

   i = -1;
   if (S == DIGIT_0) i = 0;
   if (S == DIGIT_1) i = 1;
   if (S == DIGIT_2) i = 2;
   if (S == DIGIT_3) i = 3;
   if (S == DIGIT_4) i = 4;
   if (S == DIGIT_5) i = 5;
   if (S == DIGIT_6) i = 6;
   if (S == DIGIT_7) i = 7;
   if (S == DIGIT_8) i = 8;
   if (S == DIGIT_9) i = 9;
   
   return(i);
};

static bool IsDecimalDigit(std::string S)
{
   return(InterpretDecimalDigit(S) >= 0);
};

static bool IsInt(std::string S)
{
   bool B;
   int i;
   int l;
   std::string X;
   
   if (S == "") return(FALSE);
   
   l = S.length();
   
   for (i=0; i<l; i++)
   {
      X = S.substr(i, 1);
      B = IsDecimalDigit(X);
      if (!B) return(FALSE);
   }

   try
   {
      i = std::stoi(S);
   }
   catch (const std::exception E)
   {
      return(FALSE);
   };

   return(TRUE);
};

static bool IsPositiveInt(std::string S)
{
   bool B;
   int i;
   
   B = IsInt(S);
   if (!B) return(FALSE);
   
   i = std::stoi(S);
   
   return(i >= 0);
};

Date::Date()
{
   time_t T;
   struct tm *pX;
   char *S;
   
   time(&T);
   pX = localtime(&T);
   S = asctime(pX);

   day = pX->tm_mday;
   month = pX->tm_mon + 1;
   year = pX->tm_year + 1900;
   day_of_the_week = pX->tm_wday;

   if (!IsDate(year, month, day)) throw std::invalid_argument(TEXT_BAD_DATE);
};

Date::Date(const Date& D)
{
   if (!IsDate(D.year, D.month, D.day)) throw std::invalid_argument(TEXT_BAD_DATE);

   day = D.day;
   month = D.month;
   year = D.year;
   day_of_the_week = D.day_of_the_week;
};

Date::Date(int ye, int mo, int da)
{
   if (!IsDate(ye, mo, da)) throw std::invalid_argument(TEXT_BAD_DATE);

   day = da;
   month = mo;
   year = ye;
   day_of_the_week = Compute_dotw(ye, mo, da);
};

Date::Date(const char *S)
{
   bool B;
   std::vector<std::string> V;
   std::vector<std::string>::iterator it;
   std::string T(S);
   std::string T1;
   std::string T2;
   std::string T3;
   int da;
   int mo;
   int ye;

   if (T == "") throw std::invalid_argument(TEXT_BAD_DATE);

   if (IsDateUS(S))
   {
      V = split(T, DATE_DELIMITER_US);
   }
   else
   {
      if (IsDateDE(S))
      {
         V = split(T, DATE_DELIMITER_DE);
      }
      else
      {
         throw std::invalid_argument(TEXT_BAD_DATE);
      }
   }

   if (V.size() <= 0) throw std::invalid_argument(TEXT_BAD_DATE);
   if (V.size() != 3) throw std::invalid_argument(TEXT_BAD_DATE);

   it = V.begin();

   T1 = it[0];
   B = IsPositiveInt(T1);
   if (!B) throw std::invalid_argument(TEXT_BAD_DATE);
   
   T2 = it[1];
   B = IsPositiveInt(T2);
   if (!B) throw std::invalid_argument(TEXT_BAD_DATE);

   T3 = it[2];
   B = IsPositiveInt(T3);
   if (!B) throw std::invalid_argument(TEXT_BAD_DATE);

   if (IsDateUS(S))
   {
      ye = std::stoi(T1);
      mo = std::stoi(T2);
      da = std::stoi(T3);
   }
   else
   {
      if (IsDateDE(S))
      {
         da = std::stoi(T1);
         mo = std::stoi(T2);
         ye = std::stoi(T3);
      }
      else
      {
         throw std::invalid_argument(TEXT_BAD_DATE);
      }
   }

   year = ye;
   month = mo;
   day = da;
   day_of_the_week = Compute_dotw(ye, mo, da);
};

Date::Date(const std::string S)
{
   bool B;
   std::vector<std::string> V;
   std::vector<std::string>::iterator it;
   std::string T(S);
   std::string T1;
   std::string T2;
   std::string T3;
   int da;
   int mo;
   int ye;

   if (T == "") throw std::invalid_argument(TEXT_BAD_DATE);

   if (IsDateUS(S))
   {
      V = split(T, DATE_DELIMITER_US);
   }
   else
   {
      if (IsDateDE(S))
      {
         V = split(T, DATE_DELIMITER_DE);
      }
      else
      {
         throw std::invalid_argument(TEXT_BAD_DATE);
      }
   }

   if (V.size() <= 0) throw std::invalid_argument(TEXT_BAD_DATE);
   if (V.size() != 3) throw std::invalid_argument(TEXT_BAD_DATE);

   it = V.begin();

   T1 = it[0];
   B = IsPositiveInt(T1);
   if (!B) throw std::invalid_argument(TEXT_BAD_DATE);
   
   T2 = it[1];
   B = IsPositiveInt(T2);
   if (!B) throw std::invalid_argument(TEXT_BAD_DATE);

   T3 = it[2];
   B = IsPositiveInt(T3);
   if (!B) throw std::invalid_argument(TEXT_BAD_DATE);

   if (IsDateUS(S))
   {
      ye = std::stoi(T1);
      mo = std::stoi(T2);
      da = std::stoi(T3);
   }
   else
   {
      if (IsDateDE(S))
      {
         da = std::stoi(T1);
         mo = std::stoi(T2);
         ye = std::stoi(T3);
      }
      else
      {
         throw std::invalid_argument(TEXT_BAD_DATE);
      }
   }

   year = ye;
   month = mo;
   day = da;
   day_of_the_week = Compute_dotw(ye, mo, da);
};

Date::~Date()
{
};

std::string Date::to_string_US()
{
   char buffer[256];
   std::string S;
   
   S = "";
   S += "%4.4d";
   S += DATE_DELIMITER_US;
   S += "%2.2d";
   S += DATE_DELIMITER_US;
   S += "%2.2d";

   sprintf(buffer, S.c_str(), year, month, day);
   
   std::string T(buffer);
   return(T);
};

std::string Date::to_string_DE()
{
   char buffer[256];
   std::string S;
   
   S = "";
   S += "%2.2d";
   S += DATE_DELIMITER_DE;
   S += "%2.2d";
   S += DATE_DELIMITER_DE;
   S += "%4.4d";

   sprintf(buffer, S.c_str(), day, month, year);
   
   std::string T(buffer);
   return(T);
};

int Date::DaysInMonth(int y, int m)
{
   if (m == JANUARY) return(N_JANUARY);

   if (m == FEBRUARY)
   {
      if (IsLeapYear(y))
      {
         return(N_FEBRUARY_LEAP_YEAR);
      }
      else
      {
         return(N_FEBRUARY);
      }
   }
   
   if (m == MARCH) return(N_MARCH);
   if (m == APRIL) return(N_APRIL);
   if (m == MAY) return(N_MAY);
   if (m == JUNE) return(N_JUNE);
   if (m == JULY) return(N_JULY);
   if (m == AUGUST) return(N_AUGUST);
   if (m == SEPTEMBER) return(N_SEPTEMBER);
   if (m == OCTOBER) return(N_OCTOBER);
   if (m == NOVEMBER) return(N_NOVEMBER);
   if (m == DECEMBER) return(N_DECEMBER);
   
   return(NUMERIC_NONSENSE);
};

bool Date::IsLeapYear(int y)
{
   if ((y % LEAP_YEAR_400) == 0) return(TRUE);
   if ((y % LEAP_YEAR_100) == 0) return(FALSE);
   if ((y % LEAP_YEAR_4) == 0) return(TRUE);
   
   return(FALSE);
};

bool Date::IsDate(int ye, int mo, int da)
{
   int x;

   if (ye < 1900) return(FALSE);
   if ((mo < 1) || (mo > N_MONTH)) return(FALSE);

   x = DaysInMonth(ye, mo);
   if ((da < 1) || (da > x)) return(FALSE);
   
   return(TRUE);

};

Date& Date::operator =(const Date& D)
{
   if (!IsDate(D.year, D.month, D.day)) throw std::invalid_argument(TEXT_BAD_DATE);

   day = D.day;
   month = D.month;
   year = D.year;
   day_of_the_week = D.day_of_the_week;

   return(*this);
};

// pre-fix!
Date& Date::operator ++()
{
   int x;

   if (!IsDate(year, month, day)) throw std::invalid_argument(TEXT_BAD_DATE);

   if (day_of_the_week <= FRIDAY)
   {
      day_of_the_week++;
   }
   else
   {
      // Here is (day_of_the_week == SATURDAY)!
      day_of_the_week = SUNDAY;
   }

   x = DaysInMonth(year, month);
   if (x == NUMERIC_NONSENSE) throw std::invalid_argument(TEXT_BAD_DATE);
   
   if (day < x)
   {
      day++;
      return(*this);
   }
   // Here is (day == x)!

   if (month != DECEMBER)
   {
      day = 1;
      month++;
      return(*this);
   }
   // Here is (month == DECEMBER)!

   day = 1;
   month = JANUARY;
   year++;
   return(*this);
};

// pre-fix!
Date& Date::operator --()
{
   int x;

   if (!IsDate(year, month, day)) throw std::invalid_argument(TEXT_BAD_DATE);

   if (day_of_the_week >= MONDAY)
   {
      day_of_the_week--;
   }
   else
   {
      // Here is (day_of_the_week == SUNDAY)!
      day_of_the_week = SATURDAY;
   }
   
   if (day > 1)
   {
      day--;
      return(*this);
   }
   // Here is (day == 1)!

   if (month != JANUARY)
   {
      month--;
      x = DaysInMonth(year, month);
      if (x == NUMERIC_NONSENSE) throw std::invalid_argument(TEXT_BAD_DATE);
      day = x;
      return(*this);
   }
   // Here is (month == JANUARY)!

   day = N_DECEMBER;
   month = DECEMBER;
   year--;

   if (day_of_the_week >= MONDAY)
   {
      day_of_the_week--;
   }
   else
   {
      // Here is (day_of_the_week == SUNDAY)!
      day_of_the_week = SATURDAY;
   }

   return(*this);
};

// post-fix!
Date Date::operator ++(int i)
{
   Date D(*this);
   
   ++(*this);
   
   return(D);
};

// post-fix!
Date Date::
operator --(int i)
{
   Date D(*this);

   --(*this);

   return(D);
};

// Gauss! (wikipedia)
Date Date::EasternJulianic(int ye)
{
   if (ye < 1900) throw std::invalid_argument(TEXT_BAD_DATE);

   int a = ye % 19;
   int b = ye % 4;
   int c = ye % 7;
   
   int M = 15;
   int d = (19 * a + M) % 30;
   int N = 6;
   int e = (2 * b + 4 * c + 6 * d + N) % 7;
   
   int Eastern = (22 + d + e);

   // exceptions:
   if ((d == 29) && (e == 6)) Eastern = 50;
   if ((d == 28) && (e == 6) && (((11 * M + 11) % 30) < 19)) Eastern = 49;

   if (Eastern <= N_MARCH)
   {
      Date D(ye, MARCH, Eastern);
      return(D);
   }
   else
   {
      // Here is (Eastern > N_MARCH)!
      Date D(ye, APRIL, (Eastern - N_MARCH));
      return(D);
   }
};

// Gauss! (wikipedia)
Date Date::EasternGregorianic(int ye)
{
   if (ye < 1900) throw std::invalid_argument(TEXT_BAD_DATE);

   int a = ye % 19;
   int b = ye % 4;
   int c = ye % 7;

   int k = ye / 100;
   int p = (8 * k + 13) / 25;
   int q = k / 4;
   
   int M = (15 + k - p - q) % 30;
   int d = (19 * a + M) % 30;
   int N = (4 + k - q) % 7;
   int e = (2 * b + 4 * c + 6 * d + N) % 7;
   
   int Eastern = (22 + d + e);

   // exceptions:
   if ((d == 29) && (e == 6)) Eastern = 50;
   if ((d == 28) && (e == 6) && (((11 * M + 11) % 30) < 19)) Eastern = 49;

   if (Eastern <= N_MARCH)
   {
      Date D(ye, MARCH, Eastern);
      return(D);
   }
   else
   {
      // Here is (Eastern > N_MARCH)!
      Date D(ye, APRIL, (Eastern - N_MARCH));
      return(D);
   }
};

bool Date::IsDate()
{
   return(IsDate(year, month, day));
};

bool Date::IsDate(const char *S)
{
   return((IsDateUS(S) || (IsDateDE(S))));
};

bool Date::IsDateUS(const char *S)
{
   bool B;
   std::vector<std::string> V;
   std::vector<std::string>::iterator it;
   std::string T(S);
   std::string T1;
   std::string T2;
   std::string T3;
   int da;
   int mo;
   int ye;
   
   if (T == "") return(FALSE);

   V = split(T, DATE_DELIMITER_US);
   if (V.size() <= 0) return(FALSE);
   if (V.size() != 3) return(FALSE);

   it = V.begin();

   T1 = it[0];
   B = IsPositiveInt(T1);
   if (!B) return(FALSE);
   
   T2 = it[1];
   B = IsPositiveInt(T2);
   if (!B) return(FALSE);

   T3 = it[2];
   B = IsPositiveInt(T3);
   if (!B) return(FALSE);

   ye = std::stoi(T1);
   mo = std::stoi(T2);
   da = std::stoi(T3);
   B = IsDate(ye, mo, da);
   
   return(B);
};

bool Date::IsDateDE(const char *S)
{
   bool B;
   std::vector<std::string> V;
   std::vector<std::string>::iterator it;
   std::string T(S);
   std::string T1;
   std::string T2;
   std::string T3;
   int da;
   int mo;
   int ye;
   
   if (T == "") return(FALSE);

   V = split(T, DATE_DELIMITER_DE);
   if (V.size() <= 0) return(FALSE);
   if (V.size() != 3) return(FALSE);

   it = V.begin();

   T1 = it[0];
   B = IsPositiveInt(T1);
   if (!B) return(FALSE);
   
   T2 = it[1];
   B = IsPositiveInt(T2);
   if (!B) return(FALSE);

   T3 = it[2];
   B = IsPositiveInt(T3);
   if (!B) return(FALSE);

   da = std::stoi(T1);
   mo = std::stoi(T2);
   ye = std::stoi(T3);
   B = IsDate(ye, mo, da);
   
   return(B);
};

bool Date::IsDate(const std::string S)
{
   return((IsDateUS(S) || (IsDateDE(S))));
};

bool Date::IsDateUS(const std::string S)
{
   bool B;
   std::vector<std::string> V;
   std::vector<std::string>::iterator it;
   std::string T(S);
   std::string T1;
   std::string T2;
   std::string T3;
   int da;
   int mo;
   int ye;
   
   if (T == "") return(FALSE);

   V = split(T, DATE_DELIMITER_US);
   if (V.size() <= 0) return(FALSE);
   if (V.size() != 3) return(FALSE);

   it = V.begin();

   T1 = it[0];
   B = IsPositiveInt(T1);
   if (!B) return(FALSE);
   
   T2 = it[1];
   B = IsPositiveInt(T2);
   if (!B) return(FALSE);

   T3 = it[2];
   B = IsPositiveInt(T3);
   if (!B) return(FALSE);

   ye = std::stoi(T1);
   mo = std::stoi(T2);
   da = std::stoi(T3);
   B = IsDate(ye, mo, da);
   
   return(B);
};

bool Date::IsDateDE(const std::string S)
{
   bool B;
   std::vector<std::string> V;
   std::vector<std::string>::iterator it;
   std::string T(S);
   std::string T1;
   std::string T2;
   std::string T3;
   int da;
   int mo;
   int ye;
   
   if (T == "") return(FALSE);

   V = split(T, DATE_DELIMITER_DE);
   if (V.size() <= 0) return(FALSE);
   if (V.size() != 3) return(FALSE);

   it = V.begin();

   T1 = it[0];
   B = IsPositiveInt(T1);
   if (!B) return(FALSE);
   
   T2 = it[1];
   B = IsPositiveInt(T2);
   if (!B) return(FALSE);

   T3 = it[2];
   B = IsPositiveInt(T3);
   if (!B) return(FALSE);

   da = std::stoi(T1);
   mo = std::stoi(T2);
   ye = std::stoi(T3);
   B = IsDate(ye, mo, da);
   
   return(B);
};

bool Date::operator ==(const Date& D)
{
   if (!(IsDate(year, month, day))) throw std::invalid_argument(TEXT_BAD_DATE);
   if (!(IsDate(D.year, D.month, D.day))) throw std::invalid_argument(TEXT_BAD_DATE);

   return((day == D.day) && (month == D.month) && (year == D.year));
};

bool Date::operator !=(const Date& D)
{
   return(!((*this) == D));
};

bool Date::operator <(const Date& D)
{
   if (!(IsDate(year, month, day))) throw std::invalid_argument(TEXT_BAD_DATE);
   if (!(IsDate(D.year, D.month, D.day))) throw std::invalid_argument(TEXT_BAD_DATE);

   if (year < D.year) return(TRUE);
   if (year > D.year) return(FALSE);
   // Here is (year == D.year)!
   
   if (month < D.month) return(TRUE);
   if (month > D.month) return(FALSE);
   // Here is (month == D.month)!

   if (day < D.day) return(TRUE);
   if (day > D.day) return(FALSE);
   // Here is (day == D.day)!
   
   return(FALSE);
};

bool Date::operator >(const Date& D)
{
   return((!((*this) < D)) && ((*this) != D));
};

bool Date::operator <=(const Date& D)
{
   return(!((*this) > D));
};

bool Date::operator >=(const Date& D)
{
   return(!((*this) < D));
};

int Date::operator -(const Date& D)
{
   Date X(*this);
   int n;

   if (!(IsDate(year, month, day))) throw std::invalid_argument(TEXT_BAD_DATE);
   if (!(IsDate(D.year, D.month, D.day))) throw std::invalid_argument(TEXT_BAD_DATE);

   if (X == D) return(0);

   n = 0;
   if (X < D)
   {
      while (X < D)
      {
         X++;
         n++;
      }
   }
   else
   {
      // Here is (X > D)!
      while (X > D)
      {
         X--;
         n--;
      }
   };
   
   return(n);
};

Date Date::operator +(int ndays)
{
   int i;
   Date X(*this);

   if (!(IsDate(year, month, day))) throw std::invalid_argument(TEXT_BAD_DATE);

   if (ndays == 0) return(X);

   if (ndays > 0)
   {
      for (i=0; i<ndays; i++)
      {
         X++;
      }
   }
   else
   {
      // Here is (ndays < 0)!
      for (i=0; i<-ndays; i++)
      {
         X--;
      }
   }

   return(X);
};

Date Date::operator -(int ndays)
{
   return((*this) + (-ndays));
};

Time::Time()
{
   time_t T;
   struct tm *pX;
   char *S;
   
   time(&T);
   pX = localtime(&T);
   S = asctime(pX);

   second = pX->tm_sec;
   minute = pX->tm_min;
   hour = pX->tm_hour;
};

Time::Time(const Time& T)
{
   if (!(IsTime(T.hour, T.minute, T.second))) throw std::invalid_argument(TEXT_BAD_TIME);

   hour = T.hour;
   minute = T.minute;
   second = T.second;
};

Time::Time(int ho, int mi, int se)
{
   if (!(IsTime(ho, mi, se))) throw std::invalid_argument(TEXT_BAD_TIME);

   hour = ho;
   minute = mi;
   second = se;
};

Time::Time(const char *S)
{
   bool B;
   std::vector<std::string> V;
   std::vector<std::string>::iterator it;
   std::string T(S);
   std::string T1;
   std::string T2;
   std::string T3;
   int se;
   int mi;
   int ho;
   
   if (T == "") throw std::invalid_argument(TEXT_BAD_TIME);

   V = split(T, TIME_DELIMITER);
   if (V.size() <= 0) throw std::invalid_argument(TEXT_BAD_TIME);
   if (V.size() != 3) throw std::invalid_argument(TEXT_BAD_TIME);

   it = V.begin();

   T1 = it[0];
   B = IsPositiveInt(T1);
   if (!B) throw std::invalid_argument(TEXT_BAD_TIME);
   
   T2 = it[1];
   B = IsPositiveInt(T2);
   if (!B) throw std::invalid_argument(TEXT_BAD_TIME);

   T3 = it[2];
   B = IsPositiveInt(T3);
   if (!B) throw std::invalid_argument(TEXT_BAD_TIME);

   ho = std::stoi(T1);
   mi = std::stoi(T2);
   se = std::stoi(T3);

   B = IsTime(ho, mi, se);
   if (!B) throw std::invalid_argument(TEXT_BAD_TIME);

   hour = ho;
   minute = mi;
   second = se;
};

Time::Time(const std::string S)
{
   bool B;
   std::vector<std::string> V;
   std::vector<std::string>::iterator it;
   std::string T(S);
   std::string T1;
   std::string T2;
   std::string T3;
   int se;
   int mi;
   int ho;
   
   if (T == "") throw std::invalid_argument(TEXT_BAD_TIME);

   V = split(T, TIME_DELIMITER);
   if (V.size() <= 0) throw std::invalid_argument(TEXT_BAD_TIME);
   if (V.size() != 3) throw std::invalid_argument(TEXT_BAD_TIME);

   it = V.begin();

   T1 = it[0];
   B = IsPositiveInt(T1);
   if (!B) throw std::invalid_argument(TEXT_BAD_TIME);
   
   T2 = it[1];
   B = IsPositiveInt(T2);
   if (!B) throw std::invalid_argument(TEXT_BAD_TIME);

   T3 = it[2];
   B = IsPositiveInt(T3);
   if (!B) throw std::invalid_argument(TEXT_BAD_TIME);

   ho = std::stoi(T1);
   mi = std::stoi(T2);
   se = std::stoi(T3);

   B = IsTime(ho, mi, se);
   if (!B) throw std::invalid_argument(TEXT_BAD_TIME);

   hour = ho;
   minute = mi;
   second = se;
};

Time::~Time()
{
};

std::string Time::to_string()
{
   char buffer[256];
   
   sprintf(buffer, "%2.2d:%2.2d:%2.2d", hour, minute, second);
   std::string S(buffer);
   
   return(S);
};

bool Time::IsTime(int ho, int mi, int se)
{
   if ((ho < 0) || (ho > N_HOUR)) return(FALSE);
   if ((mi < 0) || (mi > N_MINUTE)) return(FALSE);
   if ((se < 0) || (se > N_SECOND)) return(FALSE);

   if ((ho == N_HOUR) && (mi != 0) && (se != 0)) return(FALSE);

   return(TRUE);
};

bool Time::IsTime(const char *S)
{
   bool B;
   std::vector<std::string> V;
   std::vector<std::string>::iterator it;
   std::string T(S);
   std::string T1;
   std::string T2;
   std::string T3;
   int se;
   int mi;
   int ho;
   
   if (T == "") return(FALSE);

   V = split(T, TIME_DELIMITER);
   if (V.size() <= 0) return(FALSE);
   if (V.size() != 3) return(FALSE);

   it = V.begin();

   T1 = it[0];
   B = IsPositiveInt(T1);
   if (!B) return(FALSE);
   
   T2 = it[1];
   B = IsPositiveInt(T2);
   if (!B) return(FALSE);

   T3 = it[2];
   B = IsPositiveInt(T3);
   if (!B) return(FALSE);

   ho = std::stoi(T1);
   mi = std::stoi(T2);
   se = std::stoi(T3);
   B = IsTime(ho, mi, se);
   
   return(B);
};

bool Time::IsTime()
{
   return(IsTime(hour, minute, second));
};

bool Time::IsTime(const std::string S)
{
   return(Time::IsTime(S.c_str()));
};

int Time::to_seconds()
{
   int x;

   if (!(IsTime(hour, minute, second))) throw std::invalid_argument(TEXT_BAD_TIME);
   
   x = hour * N_MINUTE * N_SECOND;
   x += minute * N_SECOND;
   x += second;
   
   return(x);
};

double Time::to_minutes()
{
   double x;
   
   x = ((double) to_seconds()) / ((double) N_SECOND);
   
   return(x);
};

double Time::to_hours()
{
   double x;
   
   x = ((double) to_seconds()) / ((double) (N_SECOND * N_MINUTE));
   
   return(x);
};

Time& Time::operator =(const Time& T)
{
   if (!IsTime(T.hour, T.minute, T.second)) throw std::invalid_argument(TEXT_BAD_TIME);

   second = T.second;
   minute = T.minute;
   hour = T.hour;
   
   return(*this);
};

bool Time::operator ==(const Time& T)
{
   if (!(IsTime(hour, minute, second))) throw std::invalid_argument(TEXT_BAD_TIME);
   if (!(IsTime(T.hour, T.minute, T.second))) throw std::invalid_argument(TEXT_BAD_TIME);

   return((hour == T.hour) && (minute == T.minute) && (second == T.second));
};

bool Time::operator !=(const Time& T)
{
   return(!((*this) == T));
};

bool Time::operator <(const Time& T)
{
   if (!(IsTime(hour, minute, second))) throw std::invalid_argument(TEXT_BAD_TIME);
   if (!(IsTime(T.hour, T.minute, T.second))) throw std::invalid_argument(TEXT_BAD_TIME);

   if (hour < T.hour) return(TRUE);
   if (hour > T.hour) return(FALSE);
   // Here is (hour == T.hour)!
   
   if (minute < T.minute) return(TRUE);
   if (minute > T.minute) return(FALSE);
   // Here is (minute == T.minute)!

   if (second < T.second) return(TRUE);
   if (second > T.second) return(FALSE);
   // Here is (second == T.second)!
   
   return(FALSE);
};

bool Time::operator >(const Time& T)
{
   return((!((*this) < T)) && ((*this) != T));
};

bool Time::operator <=(const Time& T)
{
   return(!((*this) > T));
};

bool Time::operator >=(const Time& T)
{
   return(!((*this) < T));
};

DateTime::DateTime() : D(), T()
{
};

DateTime::DateTime(const Date& X, const Time& Y) : D(X), T(Y)
{
};

DateTime::DateTime(const DateTime& DT) : D(DT.D), T(DT.T)
{
};

DateTime::DateTime(int year, int month, int day, int hour, int minute, int second) :
                   D(year, month, day), T(hour, minute, second)
{
};

DateTime::DateTime(const char *S)
{
   bool B;
   int i;
   int indexDate;
   int indexTime;
   int n;
   std::vector<std::string> V;
   std::vector<std::string>::iterator it;
   std::string X(S);
   std::string T1;
   std::string T2;
   std::string U;

   if (X == "") throw std::invalid_argument(TEXT_BAD_DATETIME);
   
   V = split(X, SPACE);
   if (V.size() <= 0) throw std::invalid_argument(TEXT_BAD_DATETIME);
   
   it = V.begin();
   n = 0;
   indexDate = -1;
   indexTime = -1;
   for (i=0; i<V.size(); i++)
   {
      U = it[i];
      
      if (U != "")
      {
         n++;
         if (n == 1) indexDate = i;
         if (n == 2) indexTime = i;
      }
   }
   if (n != 2) throw std::invalid_argument(TEXT_BAD_DATETIME);
   if (indexDate < 0) throw std::invalid_argument(TEXT_BAD_DATETIME);
   if (indexTime < 0) throw std::invalid_argument(TEXT_BAD_DATETIME);

   T1 = it[indexDate];
   T2 = it[indexTime];

   B = ((Date::IsDateUS(T1.c_str()) || (Date::IsDateDE(T1.c_str()))));
   if (!B) throw std::invalid_argument(TEXT_BAD_DATETIME);
   
   B = Time::IsTime(T2.c_str());
   if (!B) throw std::invalid_argument(TEXT_BAD_DATETIME);

   D = Date(T1.c_str());
   T = Time(T2.c_str());
};

DateTime::DateTime(const std::string S)
{
   bool B;
   int i;
   int indexDate;
   int indexTime;
   int n;
   std::vector<std::string> V;
   std::vector<std::string>::iterator it;
   std::string X(S);
   std::string T1;
   std::string T2;
   std::string U;

   if (X == "") throw std::invalid_argument(TEXT_BAD_DATETIME);
   
   V = split(X, SPACE);
   if (V.size() <= 0) throw std::invalid_argument(TEXT_BAD_DATETIME);
   
   it = V.begin();
   n = 0;
   indexDate = -1;
   indexTime = -1;
   for (i=0; i<V.size(); i++)
   {
      U = it[i];
      
      if (U != "")
      {
         n++;
         if (n == 1) indexDate = i;
         if (n == 2) indexTime = i;
      }
   }
   if (n != 2) throw std::invalid_argument(TEXT_BAD_DATETIME);
   if (indexDate < 0) throw std::invalid_argument(TEXT_BAD_DATETIME);
   if (indexTime < 0) throw std::invalid_argument(TEXT_BAD_DATETIME);

   T1 = it[indexDate];
   T2 = it[indexTime];

   B = Date::IsDateUS(T1.c_str()) || Date::IsDateDE(T1.c_str());
   if (!B) throw std::invalid_argument(TEXT_BAD_DATETIME);
   
   B = Time::IsTime(T2.c_str());
   if (!B) throw std::invalid_argument(TEXT_BAD_DATETIME);

   D = Date(T1);
   T = Time(T2);
};

DateTime::~DateTime()
{
};

std::string DateTime::to_string_US()
{
   std::string S;

   S = D.to_string_US();
   S += SPACE;
   S += T.to_string();
   
   return(S);
};

std::string DateTime::to_string_DE()
{
   std::string S;

   S = D.to_string_DE();
   S += SPACE;
   S += T.to_string();
   
   return(S);
};

bool DateTime::IsDateTime(int year, int month, int day, int hour, int minute, int second)
{
   return((Date::IsDate(year, month, day)) && (Time::IsTime(hour, minute, second)));
};

bool DateTime::IsDateTime()
{
   return(IsDateTime(D.year, D.month, D.day, T.hour, T.minute, T.second));
};

bool DateTime::IsDateTime(const char *S)
{
   bool B;
   int i;
   int indexDate;
   int indexTime;
   int n;
   std::vector<std::string> V;
   std::vector<std::string>::iterator it;
   std::string X(S);
   std::string T1;
   std::string T2;
   std::string U;
   
   if (X == "") return(FALSE);
   
   V = split(X, SPACE);
   if (V.size() <= 0) return(FALSE);
   
   it = V.begin();
   n = 0;
   indexDate = -1;
   indexTime = -1;
   for (i=0; i<V.size(); i++)
   {
      U = it[i];
      
      if (U != "")
      {
         n++;
         if (n == 1) indexDate = i;
         if (n == 2) indexTime = i;
      }
   }
   if (n != 2) return(FALSE);
   if (indexDate < 0) return(FALSE);
   if (indexTime < 0) return(FALSE);

   T1 = it[indexDate];
   T2 = it[indexTime];

   B = Date::IsDateUS(T1.c_str()) || Date::IsDateDE(T1.c_str());
   if (!B) return(FALSE);

   B = Time::IsTime(T2.c_str());
   if (!B) return(FALSE);
   
   return(TRUE);
};

bool DateTime::IsDateTime(const std::string S)
{
   return(DateTime::IsDateTime((const char *) S.c_str()));
};

DateTime& DateTime::operator =(const DateTime& DT)
{
   if (!(IsDateTime(DT.D.year, DT.D.month, DT.D.day,
                    DT.T.hour, DT.T.minute, DT.T.second
                   ))) throw std::invalid_argument(TEXT_BAD_DATETIME);

   D.year = DT.D.year;
   D.month = DT.D.month;
   D.day = DT.D.day;
   
   T.hour = DT.T.hour;
   T.minute = DT.T.minute;
   T.second = DT.T.second;

   return(*this);
};

bool DateTime::operator ==(const DateTime& DT)
{
   if (!(IsDateTime(D.year, D.month, D.day,
                    T.hour, T.minute, T.second
                   ))) throw std::invalid_argument(TEXT_BAD_DATETIME);
   if (!(IsDateTime(DT.D.year, DT.D.month, DT.D.day,
                    DT.T.hour, DT.T.minute, DT.T.second
                   ))) throw std::invalid_argument(TEXT_BAD_DATETIME);

   return((D == DT.D) && (T == DT.T));
};

bool DateTime::operator !=(const DateTime& DT)
{
   return(!((*this) == DT));
};

bool DateTime::operator <(const DateTime& DT)
{
   if (!(IsDateTime(D.year, D.month, D.day,
                    T.hour, T.minute, DT.T.second
                   ))) throw std::invalid_argument(TEXT_BAD_DATETIME);
   if (!(IsDateTime(DT.D.year, DT.D.month, DT.D.day,
                    DT.T.hour, DT.T.minute, DT.T.second
                   ))) throw std::invalid_argument(TEXT_BAD_DATETIME);

   if (D < DT.D) return(TRUE);
   if (D > DT.D) return(FALSE);
   // Here is (D == DT.D)!
   
   if (T < DT.T) return(TRUE);
   if (T > DT.T) return(FALSE);
   // Here is (T == DT.T)!

   return(FALSE);
};

bool DateTime::operator >(const DateTime& DT)
{
   if (!(IsDateTime(D.year, D.month, D.day,
                    T.hour, T.minute, DT.T.second
                   ))) throw std::invalid_argument(TEXT_BAD_DATETIME);
   if (!(IsDateTime(DT.D.year, DT.D.month, DT.D.day,
                    DT.T.hour, DT.T.minute, DT.T.second
                   ))) throw std::invalid_argument(TEXT_BAD_DATETIME);

   return((!(*this < DT)) && ((*this) != DT));
};

bool DateTime::operator <=(const DateTime& DT)
{
   return(!((*this) > DT));
};

bool DateTime::operator >=(const DateTime& DT)
{
   return(!((*this) < DT));
};

int Date::Compute_dotw(int ye, int mo, int da)
{
   Date Dnow;
   Date X;
   int dotw;
   
   // Cave! X may contain illegal data.
   X.year = ye;
   X.month = mo;
   X.day = da;
   X.day_of_the_week = 0;
   
   dotw = ((Dnow - X) + Dnow.day_of_the_week) % N_WEEK;
   if (dotw < 0) dotw += N_WEEK;
   
   return(dotw);
};
